// **************************************************************************************
// el Comandante                                                    guerilla.bplaced.net
// --------------------------------------------------------------------------------------
// Information:
// Arduino esp32 --> PCF8574 I2C Interface --> Matrix keypad 4x4
// 3-Wire (SDA, SCL, !INT) I2C Matrix Keyboard Interface with the I/O expander PCF8574
// There are different types of the PCF8574 available, with different I2C address ranges
// The I2C slave address range for the PCF8574    is 0x20 - 0x27
// The I2C slave address range for the PCF8574AN  is 0x38 - 0x3F
// --------------------------------------------------------------------------------------
// Version 0 - PCF8574 / Keypad connection using the library default keypad 4x4 mode:
// I2CKeyPad Library functions: keyPad.setKeyPadMode(I2C_KEYPAD_4x4); getKey();
// PC8574 Port	   P0   P1   P2   P3   P4   P5   P6   P7
// connected to    |    |    |    |    |    |    |    |
// Keypad 4x4	     Row1 Row2 Row3 Row4 Col1 Col2 Col3 Col4 
// --------------------------------------------------------------------------------------
// Version 1 - Dollatek breakout modul with address switch:
// I2CKeyPad Library functions: keyPad.setKeyPadMode(I2C_KEYPAD_4x4P0P7); getKey();
// PC8574 Port	   P0   P1   P2   P3   P4   P5   P6   P7
// connected to    |    |    |    |    |    |    |    |
// Keypad 4x4	     Col1 Col2 Col3 Col4 Row1 Row2 Row3 Row4
// --------------------------------------------------------------------------------------
// Version 2 - Dollatek breakout modul with address jumper:
// I2CKeyPad Library functions: keyPad.setKeyPadMode(I2C_KEYPAD_4x4P7P0); getKey();
// PC8574 Port	   P7   P6   P5   P4   P3   P2   P1   P0
// connected to    |    |    |    |    |    |    |    |
// Keypad 4x4	     Col1 Col2 Col3 Col4 Row1 Row2 Row3 Row4
// --------------------------------------------------------------------------------------
// Version 3 - Custom breakout modul with address switch:
// I2CKeyPad Library functions: keyPad.setKeyPadMode(I2C_KEYPAD_4x4P0P7); getKey();
// PC8574 Port	   P0   P1   P2   P3   P4   P5   P6   P7
// connected to    |    |    |    |    |    |    |    |
// Keypad 4x4	     Col1 Col2 Col3 Col4 Row1 Row2 Row3 Row4
// --------------------------------------------------------------------------------------
// Libraries:
// TFT_eSPI   --> https://github.com/Bodmer/TFT_eSPI
// I2CKeyPad  --> https://github.com/RobTillaart/I2CKeyPad
// FT6206     --> https://github.com/adafruit/Adafruit_FT6206_Library
// --------------------------------------------------------------------------------------
// Documentations:
// Overview ESP32 mcu's   - https://socialcompare.com/en/comparison/esp8266-vs-esp32-vs-esp32-s2
// Espressif ESP32 docu's - https://www.espressif.com/en/support/documents/technical-documents
// --------------------------------------------------------------------------------------
// Device WT32-SC01       - The esp32 device WT32-SC01 can not read pixel from the TFT using HSPI
// Device WT32-SC01       - http://myosuploads3.banggood.com/products/20220111/20220111020446WT32-SC01DataSheetV3.3.pdf
// TFT Device ST7796      - https://cdn.hackaday.io/files/1625926956336128/ST7796S.pdf
// Touch Device FT6336U   - https://focuslcds.com/content/FT6236.pdf
// --------------------------------------------------------------------------------------
// Device TTGO-T-Display  - The esp32 device LILYGO-TTGO-T-Display can read pixel from the TFT
// Device TTGO-T-Display  - https://github.com/Xinyuan-LilyGO/TTGO-T-Display
// TFT Device ST7789V     - https://newhavendisplay.com/content/datasheets/ST7789V.pdf
// **************************************************************************************

#include <Math.h>                                         // Arduino Math library
#include <Wire.h>                                         // Arduino Wire library
#include <TFT_eSPI.h>                                     // TFT_espi library
#include "I2CKeyPad_mod.h"                                // I2C keypad library, modified
#include "Adafruit_FT6206.h"                              // FT6206, FT6336U - touchscreen library, modified

// ######################################################################################
// Change values in this area to match your envoirement and desires
// --------------------------------------------------------------------------------------
// I2C pins, i.e. WT32-SC01 (SDA 18,SCL 19,!INT 25), TTGO T Display (SDA 21,SCL 22,!INT 25)
#define SDA 18                                            // I2C - SDA
#define SCL 19                                            // I2C - SCL
#define INT 25                                            // !INT
//#define TFT_BL 23
const uint8_t   I2CKeyPad_Address = 0x3F;                 // i2c address (PCF8574), 0x20, 0x3F
uint8_t         mode = I2C_KEYPAD_4x4P0P7;                // 4x4 keypad matrix
// ######################################################################################

// Object(s)
TFT_eSPI tft        = TFT_eSPI();                         // 
Adafruit_FT6206 ts = Adafruit_FT6206();                   // modified > Wire.begin(18, 19)
I2CKeyPad keyPad(I2CKeyPad_Address);                      // I2C keypad library, modified

// Vars
volatile bool   keyChange = false;                        // irq var (volatile = var in RAM)
bool            seriellConnect = false;                   // seriell connection status
bool            keypadConnect = false;                    // keypad connection status
uint8_t         fontSize = 0;                             // font size, depending to screen resolution
uint32_t        width = 0;                                // screen width
uint32_t        height = 0;                               // screen height
uint32_t        fontHeight = 0;                           // font height
uint32_t        l = 0;                                    // index var
uint32_t        i = 0;                                    // index var
uint32_t        x = 0;                                    // temp var
uint32_t        y = 0;                                    // temp var
unsigned long   debounceMillis = 0;                       // manage debouncing
unsigned long   debounceMillisOld = 0;
unsigned long   debounceMillisTime = 100;
String          szTmp = "";                               // temp string

// Touch
bool      touch = false;                                  // false = touchscreen init error
uint32_t  tx;                                             // touch point x
uint32_t  ty;                                             // touch point y
uint32_t  txOld;                                          // touch point old x
uint32_t  tyOld;                                          // touch point old y

// Buttons
uint32_t        buttonAreaOrgX;                           // complete buttons area start pos x
uint32_t        buttonAreaOrgY;                           // complete buttons area start pos y
uint32_t        buttonAreaWidth;                          // complete buttons area width
uint32_t        buttonAreaHeight;                         // complete buttons area height
uint32_t        buttonAreaEndX;                           // complete buttons area end x
uint32_t        buttonAreaEndY;                           // complete buttons area end y
uint32_t        buttonPosX [16];                          // button start positions x
uint32_t        buttonPosY [16];                          // button start positions y
uint32_t        buttonEndX [16];                          // button end positions x
uint32_t        buttonEndY [16];                          // button end positions y
uint32_t        buttonCenterX [16];                       // button center x
uint32_t        buttonCenterY [16];                       // button center y
uint32_t        buttonWidth [16];                         // button width
uint32_t        buttonHeight [16];                        // button height
uint32_t        buttonGapX [16];                          // button gap y, both x sides
uint32_t        buttonGapY [16];                          // button gap y, both y sides
char            buttonLabel[18][2] = {"1", "2", "3", "A", "4", "5", "6", "B", "7", "8", "9", "C", "*", "0", "#", "D", "N", "F"}; // N = NoKey, F = Fail
uint16_t        buttonColor[16] = {0x001F, 0x001F, 0x001F, 0xF800, 0x001F, 0x001F, 0x001F, 0xF800, 0x001F, 0x001F, 0x001F, 0xF800, 0xE380, 0x001F, 0x0400, 0xF800};


// ======================================================================================
// Interrupt routine
// ======================================================================================
void IRAM_ATTR keyChanged()
{keyChange = true;}


// **************************************************************************************
// Setup
// **************************************************************************************
void setup()
{
  pinMode(TFT_BL, OUTPUT);                                // gpio - tft backlight pin
  pinMode(INT, INPUT_PULLUP);                             // gpio - irq pin

  digitalWrite(TFT_BL, 0);                                // switch backlight off

  tft.init();                                             // TFT_espi - init
  Seriell();                                              // seriell arduino monitor
  System();                                               // system
  Screen();                                               // screen
  Status();                                               // status
  Keypad();                                               // keypad

  digitalWrite(TFT_BL, 1);                                // switch backlight on
}


// **************************************************************************************
// Main Loop
// **************************************************************************************
void loop()
{
  debounceMillis = millis();
  if (debounceMillis - debounceMillisOld > debounceMillisTime)
  {
    i  = 16;
    if (keyChange || ts.touched())
    {
      if (keyChange)
      {
        i = keyPad.getKey();
        keyChange = false;                                // only after keyChange is handled reset flag
      }
      else if (ts.touched())
      {
        Touch();                                          // get the touched button
      }

      if (i < 16)
      {
        tft.setFreeFont(&FreeSansOblique12pt7b);
        tft.setTextColor(0x0000);
        for (x = 0; x < 16; x++)
        {tft.drawString(buttonLabel[x], buttonCenterX[x], buttonCenterY[x]);}
        tft.setTextColor(0xFFFF);
        tft.drawString(buttonLabel[i], buttonCenterX[i], buttonCenterY[i]);
        Serial.print("Key: "); Serial.println(buttonLabel[i]);
      }
    }
    debounceMillisOld = debounceMillis;
  }
}

// ======================================================================================
// Seriell Arduino Monitor
// ======================================================================================
void Seriell()
{
  l = 1;
  seriellConnect = true;
  Serial.begin(115200);                                   // establish arduino serial monitor connection
  while (!Serial)                                         // wait for seriell connection
  {
    if (l >= 31)                                          // wait maximal 100ms * 30 = 3000ms
    {seriellConnect = false; break;}                      // leave while loop
    delay (100);                                          // delay 100ms
    l++;                                                  // increase counter
  }
  if (seriellConnect == true)
  {Serial.print("Seriell connection: Okay in <"); Serial.print(l * 100); Serial.println("ms");}
  else
  {Serial.print("Seriell connection: Error in "); Serial.print((l-1) * 100); Serial.println("ms");}
}


// ======================================================================================
// System
// ======================================================================================
void System()
{
  // screen dimensions
  width = tft.width();                                    // native screen dimensions
  height = tft.height();
  //width = 135;                                            // in example: TTGO-T-Display
  //height = 240;

  // screen basics
  tft.setRotation(2);
  tft.setPivot(width / 2, height / 2);
  tft.setTextDatum(MC_DATUM);

  // font size depending on screen size
  if (width >= 320 || height >= 320)                      // big screen (i.e. WT32-SC01 - 480 x 320)
  {fontSize = 2; fontHeight = tft.fontHeight(2);}         // font size for big screen
  else if (width < 320 || height < 320)                   // small screen (i.e. LILYGO TTGO-T-Display - 240 x 135)
  {fontSize = 1; fontHeight = tft.fontHeight(1);}         // font size for small screen

  // init I2C
  Wire.begin(SDA, SCL);                                   // i2c - init, no error check so far :-(
  Wire.setClock(100000);

  // TOUCH
  if (!ts.begin(40))
  {touch = false; Serial.println("Touch connection  : Error");} 
  else
  {touch = true; Serial.println("Touch connection  : Okay");}

  // init keypad, attach interrupt, set keypad mode
  keypadConnect = keyPad.begin();                         // PCF8574 generates an interrupt on key press and release
  if (keypadConnect == true)                              // mode(s) defined in I2CKeyPad_mod.h
  {keyChange = false; attachInterrupt(INT, keyChanged, FALLING); keyPad.setKeyPadMode(mode); Serial.println("Keypad connection : Okay");}
  else{Serial.println("Keypad  connection: Error");}
}


// ======================================================================================
// Screen
// ======================================================================================
void Screen()
{
  tft.fillScreen(0x0000);
  tft.setTextFont(2);
  tft.setTextSize(fontSize);

  if (seriellConnect == true) tft.setTextColor(0x07FC);
  else tft.setTextColor(0xFFFF);

  tft.drawString("Arduino - esp32", width/2, height / 100.0 * 4);
  tft.setTextColor(0xFFFF);
  tft.drawString("Matrix Keypad 4*4", width/2, height / 100.0 * 11);
  tft.drawString("PCF8574 I2C Bus", width/2, height / 100.0 * 18);
  tft.drawLine(0, height/100.0 * 24, width, height/100.0 * 24, 0x8410);
}


// ======================================================================================
// Keypad
// ======================================================================================
void Keypad()
{
  // button area
  buttonAreaOrgX = width / 100.0 * 2;
  buttonAreaOrgY = height / 100.0 * 25;
  buttonAreaEndX = width / 100.0 * 98;
  buttonAreaEndY = height / 100.0 * 87;
  buttonAreaWidth = buttonAreaEndX - buttonAreaOrgX;
  buttonAreaHeight = buttonAreaEndY - buttonAreaOrgY;

  // buttons gapX, gapY, width, height
  for (x = 0; x < 16; x++)
  {
    buttonGapX[x] = width / 100.0 * 5;
    buttonGapY[x] = width / 100.0 * 5;
    buttonWidth[x] = buttonAreaWidth / 4 - buttonGapX[x];
    buttonHeight[x] = buttonAreaHeight / 4 - buttonGapY[x];
  }

  // center all buttons in area, horizontal and vertical
  i = (buttonAreaWidth - (4 * buttonWidth[0] + 3 * buttonGapX[0])) / 2.0;
  l = (buttonAreaHeight - (4 * buttonHeight[0] + 3 * buttonGapY[0])) / 2.0;

  // buttons start position, buttons end position, buttons center
  for (y = 0; y < 4; y++)
  {
    for (x = 0; x < 4; x++)
    {
      buttonPosX[x + y * 4] = i + buttonAreaOrgX + buttonGapX[x + y * 4] * x + buttonWidth[x + y * 4] * x;
      buttonPosY[x + y * 4] = l + buttonAreaOrgY + buttonGapY[x + y * 4] * y + buttonHeight[x + y * 4] * y;
      buttonEndX[x + y * 4] = buttonPosX[x + y * 4] + buttonWidth[x + y * 4];
      buttonEndY[x + y * 4] = buttonPosY[x + y * 4] + buttonHeight[x + y * 4];
      buttonCenterX[x + y * 4] = (buttonPosX[x + y * 4] + buttonEndX[x + y * 4]) / 2;
      buttonCenterY[x + y * 4] = (buttonPosY[x + y * 4] + buttonEndY[x + y * 4]) / 2 - buttonHeight[x + y * 4] / 20.0;
    }
  }

  // draw buttons
  for (i = 0; i < 16; i++)
  {
    tft.fillRoundRect(buttonPosX[i], buttonPosY[i], buttonWidth[i], buttonHeight[i], 8, buttonColor[i]);
    tft.drawRoundRect(buttonPosX[i], buttonPosY[i], buttonWidth[i], buttonHeight[i], 8, 0xFFFF);
  }

  // draw buttons text
  tft.setTextColor(0x0000);
  for (i = 0; i < 16; i++)
  {
  tft.setFreeFont(&FreeSansOblique12pt7b);
  tft.drawString(buttonLabel[i], buttonCenterX[i], buttonCenterY[i]);
  }
}


// ======================================================================================
// Status
// ======================================================================================
void Status()
{
  // Lines
  tft.drawLine(0, height/100.0 * 88, width, height/100.0 * 88, 0x8410);
  tft.drawLine(width/100.0 * 21, height/100.0 * 88, width/100.0 * 21, height, 0x8410);
  tft.drawLine(width/100.0 * 52, height/100.0 * 88, width/100.0 * 52, height, 0x8410);
  tft.drawLine(width/100.0 * 75, height/100.0 * 88, width/100.0 * 75, height, 0x8410);

  // Text
  tft.setTextFont(1);
  tft.setTextSize(fontSize);
  tft.setTextColor(0xFFFF);
  tft.drawString("I2C", width/100.0 * 10.5, height/100.0 * 92);
  tft.drawString("KEYPAD",   width/100.0 * 37.5, height/100.0 * 92);
  tft.drawString("MODE",     width/100.0 * 64.5, height/100.0 * 92);
  tft.drawString("TOUCH",    width/100.0 * 88.5, height/100.0 * 92);

  // I2C
  szTmp = String(I2CKeyPad_Address, HEX);
  szTmp.toUpperCase();
  szTmp = String("0x" + szTmp);
  if (keypadConnect == false)
  {tft.setTextColor(0xF800);}
  else
  {tft.setTextColor(0x07E0);}
   tft.drawString(szTmp, width/100.0 * 10.5, height/100.0 * 97);

  // KEYPAD
  if (keypadConnect == true)
  {tft.setTextColor(0x07E0); tft.drawString("Okay", width/100.0 * 37.5, height/100.0 * 97);}
  else
  {tft.setTextColor(0xF800); tft.drawString("Error", width/100.0 * 37.5, height/100.0 * 97);}

  // TOUCH
  if (touch == false)
  {tft.setTextColor(0xF800); tft.drawString("Error", width/100.0 * 88.5, height/100.0 * 97);} 
  else
  {tft.setTextColor(0x07E0); tft.drawString("Okay", width/100.0 * 88.5, height/100.0 * 97);}
  //{touch = false; tft.setTextColor(0xFC1F); tft.drawString("N/A", width/100.0 * 88.5, height/100.0 * 97);}

  // MODE 0xFC1F
  if (keypadConnect == true && touch == true)
  {szTmp = String(keyPad.getKeyPadMode(), DEC); tft.setTextColor(0x07E0); tft.drawString(szTmp, width/100.0 * 64.5, height/100.0 * 97);}
  else if (keypadConnect == true && touch == false)
  {szTmp = String(keyPad.getKeyPadMode(), DEC); tft.setTextColor(0xFFE0); tft.drawString(szTmp, width/100.0 * 64.5, height/100.0 * 97);}
  else if (keypadConnect == false && touch == true)
  {szTmp = String(mode, DEC); tft.setTextColor(0xFFE0); tft.drawString(szTmp, width/100.0 * 64.5, height/100.0 * 97);}
  else//if (keypadConnect == false && touch == false)
  {tft.setTextColor(0xF800); tft.drawString("None", width/100.0 * 64.5, height/100.0 * 97);}
}


// ======================================================================================
// Touch
// ======================================================================================
void Touch()
{
  TS_Point p = ts.getPoint();
  if (p.z != 0)
  {
    tx = map(p.x, 0, width, width, 0);                      // adjust touch point for rotation = 2
    ty = map(p.y, 0, height, height, 0);

    if (tx != txOld && ty != tyOld)
    {
      if (ty >= height/100.0 * 93)
      {ESP.restart();}

      for (l = 0; l < 16; l++)
      {
        if (tx > buttonPosX[l] && tx < buttonEndX[l] && ty > buttonPosY[l] && ty < buttonEndY[l])
        {i = l; break;}
      }
    
      txOld = tx;
      tyOld = ty;
    }
  }
}
